#include "textchecker.h"
#include <unordered_map>
#include <cwctype>
#ifdef SORT
#include "quicksort_for_vector_wstring.h"
#include <algorithm>
#endif

textChecker::textChecker() {}
textChecker::textChecker(std::wstring const &a):text(a){}
textChecker::textChecker(textChecker const &a):text(a.text){}
void textChecker::set(std::wstring const &a){text=a;}
bool operator==(const textChecker& num1, const textChecker& num2){return (num1.text==num2.text);}
void textChecker::reset(){text=L"";}
const std::wstring &textChecker::towstring(){return text;}
bool textChecker::cyrtolat() {
    static const std::unordered_map<wchar_t, std::wstring> translit_map = {
        {L'А', L"A"}, {L'а', L"a"}, {L'Б', L"B"}, {L'б', L"b"}, {L'В', L"V"}, {L'в', L"v"},
        {L'Г', L"H"}, {L'г', L"h"}, {L'Ґ', L"G"}, {L'ґ', L"g"}, {L'Д', L"D"}, {L'д', L"d"},
        {L'Е', L"E"}, {L'е', L"e"}, {L'Є', L"Ye"}, {L'є', L"ye"}, {L'Ж', L"Zh"}, {L'ж', L"zh"},
        {L'З', L"Z"}, {L'з', L"z"}, {L'Й', L"Y"}, {L'й', L"y"}, {L'І', L"I"}, {L'і', L"i"},
        {L'Ї', L"Yi"}, {L'ї', L"yi"}, {L'И', L"Y"}, {L'и', L"y"}, {L'К', L"K"}, {L'к', L"k"},
        {L'Л', L"L"}, {L'л', L"l"}, {L'М', L"M"}, {L'м', L"m"}, {L'Н', L"N"}, {L'н', L"n"},
        {L'О', L"O"}, {L'о', L"o"}, {L'П', L"P"}, {L'п', L"p"}, {L'Р', L"R"}, {L'р', L"r"},
        {L'С', L"S"}, {L'с', L"s"}, {L'Т', L"T"}, {L'т', L"t"}, {L'У', L"U"}, {L'у', L"u"},
        {L'Ф', L"F"}, {L'ф', L"f"}, {L'Х', L"Kh"}, {L'х', L"kh"}, {L'Ц', L"Ts"}, {L'ц', L"ts"},
        {L'Ч', L"Ch"}, {L'ч', L"ch"}, {L'Ш', L"Sh"}, {L'ш', L"sh"}, {L'Щ', L"Shch"}, {L'щ', L"shch"},
        {L'Ь', L"'"}, {L'ь', L"'"}, {L'Ю', L"Yu"}, {L'ю', L"yu"}, {L'Я', L"Ya"}, {L'я', L"ya"}
    };
    
    std::wstring original_text = text;
    std::wstring result;
    
    for (const wchar_t &c : text)
        if (translit_map.find(c) != translit_map.end()) //Проста заміна символу на його транслітерацію
            result += translit_map.at(c);
        else
            result += c;
    bool changed = (original_text != result);
    text = result;
    return changed ? 1 : 0;
}
bool textChecker::lattocyr() {
    static const std::unordered_map<std::wstring, wchar_t> translit_map = {
        {L"Ye", L'Є'}, {L"ye", L'є'}, {L"Yi", L'Ї'}, {L"yi", L'ї'},
        {L"Yu", L'Ю'}, {L"yu", L'ю'}, {L"Ya", L'Я'}, {L"ya", L'я'},
        {L"Zh", L'Ж'}, {L"zh", L'ж'}, {L"Kh", L'Х'}, {L"kh", L'х'},
        {L"Shch", L'Щ'}, {L"shch", L'щ'}, {L"Sh", L'Ш'}, {L"sh", L'ш'},
        {L"Ts", L'Ц'}, {L"ts", L'ц'}, {L"Ch", L'Ч'}, {L"ch", L'ч'}
    };
    
    static const std::unordered_map<wchar_t, wchar_t> single_map = {
        {L'A', L'А'}, {L'a', L'а'}, {L'B', L'Б'}, {L'b', L'б'}, {L'V', L'В'}, {L'v', L'в'},
        {L'H', L'Г'}, {L'h', L'г'}, {L'G', L'Ґ'}, {L'g', L'ґ'}, {L'D', L'Д'}, {L'd', L'д'},
        {L'E', L'Е'}, {L'e', L'е'}, {L'Z', L'З'}, {L'z', L'з'}, {L'I', L'І'}, {L'i', L'і'},
        {L'Y', L'И'}, {L'y', L'и'}, {L'K', L'К'}, {L'k', L'к'}, {L'L', L'Л'}, {L'l', L'л'},
        {L'M', L'М'}, {L'm', L'м'}, {L'N', L'Н'}, {L'n', L'н'}, {L'O', L'О'}, {L'o', L'о'},
        {L'P', L'П'}, {L'p', L'п'}, {L'R', L'Р'}, {L'r', L'р'}, {L'S', L'С'}, {L's', L'с'},
        {L'T', L'Т'}, {L't', L'т'}, {L'U', L'У'}, {L'u', L'у'}, {L'F', L'Ф'}, {L'f', L'ф'},
        {L'\'', L'ь'}, {L'Q', L'К'}, {L'q', L'к'}, {L'W', L'В'}, {L'w', L'в'},
        {L'J', L'Й'}, {L'j', L'й'}, {L'X', L'Х'}, {L'x', L'х'}, {L'C', L'Ц'}, {L'c', L'ц'}
    };
    
    std::wstring original_text = text;
    std::wstring result;
    size_t i = 0;
    
    while (i < text.length()) {
        std::wstring next_two = (i + 1 < text.length()) ? text.substr(i, 2) : L"";
        std::wstring next_four = (i + 3 < text.length()) ? text.substr(i, 4) : L"";
        
        if (translit_map.find(next_four) != translit_map.end()) {//Найбільший пріоритет для найдовших транслітерацій
            result += translit_map.at(next_four);
            i += 4;
        } else if (translit_map.find(next_two) != translit_map.end()) {
            result += translit_map.at(next_two);
            i += 2;
        } else if (single_map.find(text[i]) != single_map.end()) {
            result += single_map.at(text[i]);
            i++;
        } else {
            result += text[i];
            i++;
        }
    }
    
    bool changed = (original_text != result);
    text = result;
    return changed ? 1 : 0;
}

const std::array<float, 4> textChecker::statc() const {
    unsigned int cyr=0,lat=0,num=0,unk=0,all=0;
    for (const wchar_t &c : this->text){all++;
        if (std::isdigit(c)){num++;continue;}
        else if ((c>=L'a'&&c<=L'z')||(c>=L'A'&&c<=L'Z')){lat++;continue;}
        else if (c>=L'Ѐ'&&c<=L'ӿ')cyr++;
        else unk++;
    }
    std::array<float, 4> stat;
    if(all){
    stat[0] = static_cast<float>(cyr) / all;
    stat[1] = static_cast<float>(lat) / all;
    stat[2] = static_cast<float>(num) / all;
    stat[3] = static_cast<float>(unk) / all;}
    else{
        stat[0] = 0;
        stat[1] = 0;
        stat[2] = 0;
        stat[3] = 0;
    }return stat;}
const std::array<unsigned short, 10> textChecker::statn() const {
    std::array<unsigned short, 10> digitStats = {0,0,0,0,0,0,0,0,0,0};
    for (wchar_t c : this->text)if (c >= L'0' && c <= L'9')digitStats[c - L'0']++;
    return digitStats;
}
std::ostream& operator<<(std::ostream& os, const textChecker &text){
    os << text;
    return os;
}
std::wistream& operator>>(std::wistream& is, textChecker& text) {
    std::wstring temp;
    is >> temp;
    if (!temp.empty() && temp.back() != L' ') {
        if (!text.text.empty() && text.text.back() != L' ') {
            text.set(text.text + L' ' + temp);
        } else {
            text.set(text.text + temp);
        }
    }
    return is;
}

unsigned int textChecker::swap(wchar_t from,wchar_t to){
    unsigned int count=0;
    for (wchar_t& ch : text) {
        if (ch == from) {
            ch = to;
            count++;
        }
    }
    return count;
}
unsigned int textChecker::swap(const char from,const char to) {return swap(static_cast<wchar_t>(from),static_cast<wchar_t>(to));}
unsigned int textChecker::swap(const char from,const wchar_t to) {return swap(static_cast<wchar_t>(from),to);}
unsigned int textChecker::swap(const wchar_t from,const char to) {return swap(from,static_cast<wchar_t>(to));}
const unsigned int textChecker::sentences() {
    unsigned int count = 0;
    for (size_t i = 0; i < text.size(); ++i)
        if (isend(text[i])) {
            //if (i >= 1 && text[i - 1] == L'.'|| text[i - 1] == L'!' || text[i - 1] == L'?')continue; В процесі реалізації це стало непотрібним
            skip;
            count++;
        }
    return count;
}

const std::wstring textChecker::longestSentence() {
    size_t maxLength = 0;
    size_t currentStart = 0;
    size_t longestStart = 0;
    for (size_t i = 0; i < text.size(); ++i) {
        if (isend(text[i])) {
            skip;
            size_t currentLength = i - currentStart + 1;          //Додавання всюди зумовлено підрахунком з нуля. 
            if (currentLength > maxLength) {                      //Наприклад, речення
                maxLength = currentLength;                        //А. встановить i в 1 та substr відкусить відповідно "А"
                longestStart = currentStart;                      //А... встановить i в 3 за допомогою інструкцій skip,
            }currentStart = i + 1;                                //буде відкушено "A..". Як бачимо, додавання одиниці
        }                                                         //актуально та необхідно за будь-якої стрічки
    }
    return text.substr(longestStart, maxLength);
}

const std::wstring textChecker::shortestSentence() {// Буквально те саме, що й з longestSentence, але зі зміненим знаками нерівності
    size_t minLength = std::wstring::npos;
    size_t currentStart = 0;
    size_t shortestStart = 0;
    for (size_t i = 0; i < text.size(); ++i) {
        if (isend(text[i])) {
            skip;
            size_t currentLength = i - currentStart + 1;
            if (currentLength < minLength) {
                minLength = currentLength;
                shortestStart = currentStart;
            }currentStart = i + 1;
        }
    }return text.substr(shortestStart+1, minLength);
}

const std::
#ifdef SORT
    vector
#else
    set
#endif
<std::wstring> textChecker::capSort() {
    std::
    #ifdef SORT
        vector
    #else
        set
    #endif
        <std::wstring> words;
    std::wstring currentWord;
    
    for (size_t i = 0; i < text.size(); ++i) {
        wchar_t currentChar = text[i];
        if (std::iswalpha(currentChar))
            currentWord += currentChar;
        else
            if (!currentWord.empty()) {
                if (std::iswupper(currentWord[0]))
                  words.
#ifdef SORT
                  push_back
#else
                    insert
#endif
                        (currentWord);
                currentWord.clear();
            }
    }
#ifdef SORT
    quicksort(words, 0, words.size() - 1);
    words.erase(unique(words.begin(),words.end()),words.end());    
#endif
    return words;
}

const bool textChecker::empty(){return text.empty();}
